/***************************************************************

 UciFilter.cpp
 
 Filtert die UCI-Kommunikation zwischen Arena und Engine
 
***************************************************************/

// assert weg definieren
#define NDEBUG


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <windows.h>


#ifdef _MSC_VER
typedef __int64 Int_64;
#else
typedef long long Int_64;
#endif

typedef struct
{
	// Dateihandles
	HANDLE	hVonEngine;
	HANDLE	hNachEngine;
	HANDLE	hVonGui;
	HANDLE	hNachGui;
	HANDLE	hEngineNachFilter;
	
	// Engine-Prozehandle
	HANDLE	hEngine;

	// Threadhandle
	HANDLE	hThread;

	// Zustand
	volatile BOOL	bEngineRechnet;
	volatile BOOL	bQuit;
	volatile Int_64	ZeitCurrmove;
	volatile Int_64	ZeitInfo;
	
	// Optionen
	volatile long	ValueCurrmove;
	volatile long	ValueInfo;
	volatile long	ValueQuit;
	volatile BOOL	bFilterNonUci;
	volatile BOOL	bFilterStop;
}
FilterDaten;

typedef struct
{
	HANDLE	hDatei;			// Dateihandle
	char* 	pPuffer;		// Datenpuffer
	size_t	Groesse;		// Gre des Datenpuffers - 1, 1 Extrazeichen fr Stringendemarke
	size_t	AnzInhalt;		// Anzahl Zeichen im Puffer ohne Endemarke
	size_t	AnzVerarbeitet;	// Anzahl Zeichen, die bereits verarbeitet wurden
} Lesepuffer;


void         FatalerFehler             ( const char* meldung, HANDLE hEngine );
Int_64       Jetzt                     ( void );
void         HandleDurchPipeErsetzen   ( DWORD EinOderAus, HANDLE* pMeinHandle, HANDLE* pEngineHandle, HANDLE* pOriginal );
HANDLE       EngineStarten             ( int argc, char* argv[] );
void         WriteOption               ( HANDLE h, const char* str );
char*        StringLesen               ( Lesepuffer* pLP, HANDLE hEngine );
void         GuiNachEngine             ( FilterDaten* pFilter );
DWORD WINAPI EngineNachGui             ( LPVOID pParam );
void         EngineNachGuiThreadStarten( FilterDaten *pFilter );



void FatalerFehler( const char* meldung, HANDLE hEngine )
{
	DWORD LastError = GetLastError();
	fprintf( stderr, "info string UciFilter: FEHLER %ld; %s\n", (long)LastError, meldung );
	fflush( stderr );
	
	if ( hEngine != NULL )
		TerminateThread( hEngine, 0 );
	
	exit(0);
}


// Systemzeit in Millisekunden seit wann auch immer
Int_64 Jetzt( void )
{
	FILETIME ft;
	LARGE_INTEGER li;

	GetSystemTimeAsFileTime( &ft );

	li.LowPart		= ft.dwLowDateTime;
	li.HighPart		= ft.dwHighDateTime;

	return li.QuadPart / 10000;
}


// HandleDurchPipeErsetzen()
//
// - Ersetzt STD_INPUT_HANDLE oder STD_OUTPUT_HANDLE durch ein
//   vererbbares Ende einer Pipe, damit dieses an einen Kindprozess
//   vererbt werden kann.
// - Liefert ein Handle auf das andere Ende der Pipe zurck,
//   ber den Daten mit dem Kindprozess ausgetauscht werden knnen.
// - Liefert das Original-Handle zurck, damit nichts verloren geht.
void HandleDurchPipeErsetzen( DWORD EinOderAus, HANDLE* pMeinHandle, HANDLE* pEngineHandle, HANDLE* pOriginal )
{
	SECURITY_ATTRIBUTES SecAttr;
	HANDLE              hRein, hRaus, hDoppel, hWasBinIch;
	HANDLE				*phMeins, *phEngineIhrs;

	// SECURITY_ATTRIBUTES initialisieren
	memset( &SecAttr, 0, sizeof( SecAttr ) );
	SecAttr.nLength        = sizeof( SecAttr );
	SecAttr.bInheritHandle = TRUE;
	
	// Die Original Standardein- bzw. -ausgabe sichern
	*pOriginal = GetStdHandle( EinOderAus );

	// Pipe erzeugen
	if ( !CreatePipe( &hRein, &hRaus, &SecAttr, 0 ) )
		FatalerFehler( "CreatePipe geht nicht\n", NULL );

	// Was ist fr mich, was fr die Engine?
	if ( EinOderAus == STD_OUTPUT_HANDLE )
	{
		phMeins      = &hRein;
  		phEngineIhrs = &hRaus;
	}
	else
	{
		phMeins      = &hRaus;
  		phEngineIhrs = &hRein;
	}
	
	// Standardein- bzw. -ausgabe umbiegen
	if ( !SetStdHandle( EinOderAus, *phEngineIhrs ) )
		FatalerFehler( "Ein- oder Ausgabe umbiegen geht nicht", NULL );

	// Welches Schweinderl htten Sie denn gerne? ;-)
	hWasBinIch = GetCurrentProcess();
	
	// Jetzt machen wir das eigene Handle durch Duplizieren nicht vererbbar
    if ( !DuplicateHandle( hWasBinIch, *phMeins,
                           hWasBinIch, &hDoppel,
                           0, FALSE, DUPLICATE_SAME_ACCESS
                         )
       )
       	FatalerFehler( "Doppeln fehlgeschlagen", NULL );

	// Weg mit dem vererbbaren Original
    CloseHandle( *phMeins );
    
    // Handles auf beide Enden der Pipe zurck geben
	*pMeinHandle   = hDoppel;
	*pEngineHandle = *phEngineIhrs;
}

HANDLE EngineStarten( int argc, char* argv[] )
{
	PROCESS_INFORMATION	ProcessInfo;
	STARTUPINFO			StartInfo;
	char				Kommandozeile[10000];
	char				Dir[_MAX_DIR];
	char				DirDrive[_MAX_PATH];
	int          		i;
	size_t				laenge;

	// Strukturen initialisieren
	memset( &ProcessInfo, 0, sizeof( ProcessInfo ) );
	memset( &StartInfo,   0, sizeof( StartInfo   ) );
	StartInfo.cb = sizeof( StartInfo );

	// Kommandozeile zusammensetzen
	Kommandozeile[0] = '\0';
	laenge           = 0;
	for ( i=1; i<argc; i++ )
	{
		if ( laenge + strlen( argv[i] ) + 1 >= sizeof( Kommandozeile ) )
			FatalerFehler( "Die Kommandozeile ist zu lang", NULL );
		if ( i > 1 )
			strcat( Kommandozeile, " " );
		strcat( Kommandozeile, argv[i] );
	}

	// Verzeichnis fr Engine ermitteln
	_splitpath( argv[1], DirDrive, Dir, NULL, NULL );
	strcat( DirDrive, Dir );

	// los geht's
	if ( !CreateProcess( argv[1],
                         Kommandozeile,
						 NULL,			// security attributes
						 NULL,			// thread security attributes
						 TRUE,			// handles vererben
						 0,				// flags
						 NULL,			// Umgebung
						 DirDrive,		// Engine-Verzeichnis
						 &StartInfo,	// StartInfo
						 &ProcessInfo	// Prozessinformationen
	    				)
	   )
		FatalerFehler( "Engineprozess lsst sich nicht starten", NULL );
		
	return ProcessInfo.hProcess;
}

char* StringLesen( Lesepuffer* pLP, HANDLE hEngine )
{
	char *pAnfang, *pEnde;
	DWORD AnzGelesen;

	// Ist noch was im Puffer
	if ( pLP->AnzInhalt > pLP->AnzVerarbeitet )
	{
		// Ist es ein vollstndiger String?
		pAnfang = pLP->pPuffer + pLP->AnzVerarbeitet;
		pEnde   = strchr( pLP->pPuffer + pLP->AnzVerarbeitet, '\n' );
		if ( pEnde != NULL )
		{
			// Den String schon mal als verarbeitet kennzeichnen
			pLP->AnzVerarbeitet = pEnde - pLP->pPuffer + 1;
			assert( pLP->AnzInhalt >= pLP->AnzVerarbeitet );

			// String terminieren
			assert( isspace( *pEnde ) );
			while ( isspace( *pEnde ) && pEnde >= pAnfang )
      			*pEnde-- = '\0';

			// fertig
			return pAnfang;
		}
		else
		{
  			// es ist etwas im Puffer, aber kein vollstndiger String
			if ( pLP->AnzVerarbeitet > 0 )
			{
	  			// Daten nach vorne verschieben, evtl. berlappend!
  				memmove( pLP->pPuffer, pAnfang, pLP->AnzInhalt-pLP->AnzVerarbeitet+1 );
  				pLP->AnzInhalt     -= pLP->AnzVerarbeitet;
  				pLP->AnzVerarbeitet = 0;
			}
		}
	}
	else
	{
		// nichts drin, auf Anfang zurcksetzen
		pLP->AnzInhalt      = 0;
		pLP->AnzVerarbeitet = 0;
	}
	
	do	// Datei (pipe) lesen, bis ein vollstndiger String drin ist
	{
		// Datei lesen
		if ( !ReadFile( pLP->hDatei, pLP->pPuffer+pLP->AnzInhalt, pLP->Groesse-pLP->AnzInhalt, &AnzGelesen, NULL ) )
			FatalerFehler( "Lesefehler", hEngine );

  		pLP->AnzInhalt += AnzGelesen;
  		assert( pLP->AnzInhalt <= pLP->Groesse );
  		
  		// Puffer nullterminieren
  		pLP->pPuffer[pLP->AnzInhalt] = '\0';

		pAnfang = pLP->pPuffer + pLP->AnzVerarbeitet;
		pEnde   = strchr( pAnfang, '\n' );
		if ( pEnde == NULL )
		{
			// immer noch kein vollstndiger String
			
			// Puffer voll?
			if ( pLP->AnzInhalt == pLP->Groesse )
			{
				// P-Verlngerung ;-)
				pLP->Groesse *= 2;
				pLP->pPuffer  = realloc( pLP->pPuffer, pLP->Groesse+1 );
				if ( pLP->pPuffer == NULL )
        			FatalerFehler( "Nicht gengend Speicher", hEngine );
			}
		}
	}
	while ( pEnde == NULL );

	// String liefern wie oben

	// Den String schon mal als verarbeitet kennzeichnen
	pLP->AnzVerarbeitet = pEnde - pLP->pPuffer + 1;
	assert( pLP->AnzInhalt >= pLP->AnzVerarbeitet );

	// String terminieren
	assert( isspace( *pEnde ) );
	while ( isspace( *pEnde ) && pEnde >= pAnfang )
		*pEnde-- = '\0';

	// fertig
	return pAnfang;
}

void GuiNachEngine( FilterDaten* pFilter )
{
	Lesepuffer lp;
	char       Kommando[16];
	char      *str;
	char      *strName;
	char      *strOption;
	char      *strValue;
	size_t     len;
	DWORD      AnzGeschrieben;
	
	// Lesepuffer initialisieren
	lp.hDatei         = pFilter->hVonGui;
	lp.Groesse        = 10000;
	lp.pPuffer        = malloc( lp.Groesse+1 );
	lp.AnzInhalt      = 0;
	lp.AnzVerarbeitet = 0;
	if ( lp.pPuffer == NULL )
		FatalerFehler( "Nicht gengend Speicherplatz", pFilter->hEngine );
	
	while ( TRUE ) // Ausstieg innen
	{
		// lesen
  		str = StringLesen( &lp, pFilter->hEngine );
  		len = strlen( str );
  		
  		// Kommando extrahieren
  		sscanf( str, "%15s", Kommando );
  		
  		if ( strcmp( Kommando, "setoption" ) == 0 )
  		{
  			// Wir interessieren uns nur fr Optionen mit '@UciFilter' drin
  			strName   = strstr( str, " name " );
  			strOption = strstr( str, "@UciFilter " );
  			strValue  = strstr( str, " value " );
  			if (   strName   != NULL		// 'name' muss sein
  				&& strOption != NULL		// nur '@UciFilter' interessiert
  				&& strValue  != NULL		// 'value' muss auch sein
  				&& strName   < strOption	// Reihenfolge testen
  				&& strOption < strValue
			   )
  			{
  				if ( strncmp( strOption-9, " TimeQuit", 9 ) == 0 )
  				{
  					sscanf( strValue+7, "%ld", &pFilter->ValueQuit );
  					continue;
  				}
  				else if ( strncmp( strOption-13, " TimeCurrmove", 13 ) == 0 )
  				{
  					sscanf( strValue+7, "%ld", &pFilter->ValueCurrmove );
  					continue;
  				}
  				else if ( strncmp( strOption-9, " TimeInfo", 9 ) == 0 )
  				{
  					sscanf( strValue+7, "%ld", &pFilter->ValueInfo );
      				continue;
  				}
  				else if ( strncmp( strOption-11, " FilterStop", 11 ) == 0 )
  				{
           			strValue += 7;
           			while ( *strValue != '\0' && isspace( *strValue ) )
           				strValue++;
           			pFilter->bFilterStop = memcmp( strValue, "true", 4 ) == 0;
           			continue;
  				}
  				else if ( strncmp( strOption-13, " FilterNonUci", 13 ) == 0 )
  				{
           			strValue += 7;
           			while ( *strValue != '\0' && isspace( *strValue ) )
           				strValue++;
           			pFilter->bFilterNonUci = memcmp( strValue, "true", 4 ) == 0;
      				continue;
  				}
			}
  		}
  		else if ( strcmp( Kommando, "go" ) == 0 )
  		{
  			// fr stop-Filter
  			pFilter->bEngineRechnet = TRUE;
  			
  			// Zeiten fr info-Filter
     		pFilter->ZeitCurrmove   = Jetzt();
       		pFilter->ZeitInfo       = pFilter->ZeitCurrmove;
     		pFilter->ZeitCurrmove  += pFilter->ValueCurrmove;
       		pFilter->ZeitInfo      += pFilter->ValueInfo;
  		}
  		else if ( strcmp( Kommando, "stop" ) == 0 )
  		{
  			// berflssiges stop-Kommaodo ausfiltern
  			if ( pFilter->bFilterStop && !pFilter->bEngineRechnet )
  				continue;
  		}
  		else if ( strcmp( Kommando, "quit" ) == 0 )
  		{
  			// dem EngineNachGui-Thread zeigen, dass er sich beenden soll
			pFilter->bQuit = TRUE;
			
			// dem EngineNachGui-Thread eine Eingabe liefern, damit er aufwacht
			WriteFile( pFilter->hEngineNachFilter, "\n", 1, &AnzGeschrieben, NULL );
	        FlushFileBuffers( pFilter->hEngineNachFilter );
	        
	        // Keine Fehlerprfung!
	        // Wenn es nicht geht, verschwindet der Thread mit diesem Prozess
  		}
  		
  		// Kommando durchreichen
  		str[len] = '\n';
		if (   !WriteFile( pFilter->hNachEngine, str, len+1, &AnzGeschrieben, NULL )
			|| AnzGeschrieben != len+1
           )
           	FatalerFehler( "Schreibfehler", pFilter->hEngine );
        FlushFileBuffers( pFilter->hNachEngine );

		// Nachbehandlung "quit"
		if ( strcmp( Kommando, "quit" ) == 0 )
		{
			if ( WaitForSingleObject( pFilter->hEngine, pFilter->ValueQuit ) != WAIT_OBJECT_0 )
			{
				// Auf die harte Tour beenden
				TerminateProcess( pFilter->hEngine, 0 );
			}
			
			// Wenn durch den Engineprozess die Pipe weg ist,
   			// sollte auch der EngineNachGui-Thread terminieren
   			WaitForSingleObject( pFilter->hThread, 300 );
   			
   			// Fertig
   			return;
		}
	}
}

void WriteOption( HANDLE h, const char* str )
{
	DWORD AnzGeschrieben;
	
	WriteFile( h, str, strlen( str ), &AnzGeschrieben, NULL );
}

DWORD WINAPI EngineNachGui( LPVOID pParam )
{
	FilterDaten *pFilter;
	Lesepuffer   lp;
	char         Kommando[16];
	char        *str;
	size_t       len;
	DWORD        AnzGeschrieben;

	// Thread-Parameter ist vom Typ FilterDaten
	pFilter = (FilterDaten*)pParam;

	// Lesepuffer initialisieren
	lp.hDatei         = pFilter->hVonEngine;
	lp.Groesse        = 10000;
	lp.pPuffer        = malloc( lp.Groesse+1 );
	lp.AnzInhalt      = 0;
	lp.AnzVerarbeitet = 0;
	if ( lp.pPuffer == NULL )
		FatalerFehler( "Nicht gengend Speicherplatz", pFilter->hEngine );

	while ( TRUE ) // Ausstieg innen
	{
		// lesen
  		str = StringLesen( &lp, pFilter->hEngine );
  		len = strlen( str );

		// Falls quit-Kommando empfangen wurde, wird nichts mehr weiter geleitet.
		// Dieser Thread endet dann sofort.
		if ( pFilter->bQuit )
			return 0;

  		// Kommando extrahieren
  		sscanf( str, "%15s", Kommando );

  		if ( strcmp( Kommando, "uciok" ) == 0 )
  		{
  			WriteOption( pFilter->hNachGui, "option name TimeCurrmove@UciFilter type spin default 1500 min 0 max 5000\n" );
  			WriteOption( pFilter->hNachGui, "option name TimeInfo@UciFilter type spin default 300 min 0 max 5000\n" );
  			WriteOption( pFilter->hNachGui, "option name TimeQuit@UciFilter type spin default 1000 min 200 max 1800\n" );
  			WriteOption( pFilter->hNachGui, "option name RemoveNonUci@UciFilter type check default true\n" );
  			WriteOption( pFilter->hNachGui, "option name RemoveStop@UciFilter type check default true\n" );
  		}
  		else if ( strcmp( Kommando, "bestmove" ) == 0 )
  		{
  			pFilter->bEngineRechnet = FALSE;
  		}
  		else if ( strcmp( Kommando, "info" ) == 0 )
  		{
  			if ( strstr( str, " currmove " ) != NULL )
  			{
  				// info mit 'currmove' filtern
				if ( Jetzt() < pFilter->ZeitCurrmove )
	  				continue;
			}
			else
  			{
  				// sonstige info-Strings filtern
				if ( Jetzt() < pFilter->ZeitInfo )
	  				continue;
			}
  		}
  		else if (  pFilter->bFilterNonUci
//				&& strcmp( Kommando, "uciok"          ) == 0
//				&& strcmp( Kommando, "bestmove"       ) == 0
//				&& strcmp( Kommando, "info"           ) == 0
				&& strcmp( Kommando, "id"             ) == 0
				&& strcmp( Kommando, "readyok"        ) == 0
				&& strcmp( Kommando, "copyprotection" ) == 0
				&& strcmp( Kommando, "option"         ) == 0
				)
		{
  			continue;
		}

  		// Kommando durchreichen
  		str[len] = '\n';
		if (   !WriteFile( pFilter->hNachGui, str, len+1, &AnzGeschrieben, NULL )
			|| AnzGeschrieben != len+1
           )
           	FatalerFehler( "Schreibfehler", pFilter->hEngine );

		// bestmove so schnell wie mglich durchreichen
		if ( !pFilter->bEngineRechnet )
	        FlushFileBuffers( pFilter->hNachGui );
	}
}


void EngineNachGuiThreadStarten( FilterDaten *pFilter )
{
	ULONG ThreadID;
	pFilter->hThread = CreateThread( NULL, 256000, EngineNachGui, pFilter, 0, &ThreadID );
}

int main( int argc, char* argv[] )
{
	FilterDaten Filter;
	HANDLE	hEngineLesen;

	// Erstes Kommandozeilenargument ist die Engine
	if ( argc < 2 )
		FatalerFehler( "Keine Engine angegeben", NULL );

	// Filter nitialisieren
	memset( &Filter, 0, sizeof( Filter ) );
	Filter.ValueCurrmove = 1500;
	Filter.ValueInfo     =  300;
	Filter.ValueQuit     = 1000;
	Filter.bFilterStop   = TRUE;
 	Filter.bFilterNonUci = TRUE;
	
	// Standardein- und -ausgabe durch Pipes ersetzen
	HandleDurchPipeErsetzen( STD_OUTPUT_HANDLE, &Filter.hVonEngine,  &Filter.hEngineNachFilter, &Filter.hNachGui );
	HandleDurchPipeErsetzen( STD_INPUT_HANDLE,  &Filter.hNachEngine, &hEngineLesen,             &Filter.hVonGui  );

	// Engine Starte
	Filter.hEngine = EngineStarten( argc, argv );

 	// Standardein- und -ausgabe wieder durch das Original ersetzen
	if (   !SetStdHandle( STD_INPUT_HANDLE,  Filter.hVonGui  )
		|| !SetStdHandle( STD_OUTPUT_HANDLE, Filter.hNachGui )
       )
		FatalerFehler( "Wiederherstellung von Standardein- oder -ausgabe fehlgeschlagen", Filter.hEngine );

	// Das Pipe-Ende von dem die Engine liest in diesem Prozess schlieen
	if ( !CloseHandle( hEngineLesen ) )
		FatalerFehler( "Schlieen des lesenden Pipe-Endes bei Engine fehlgeschlagen", Filter.hEngine );

	// Das Pipe-Ende in das die Engine schreibt bleibt hier offen! Es wird dazu
	// benutzt, um den EngineNachGui-Thread aufzuwecken, wenn der beim Lesen
	// aus der Pipe blockiert ist und beendet werden soll.

	// Jede Richtung bekommt ihren eigenen Thread.
	// Die Rckrichtung wird extra gestartet.
	EngineNachGuiThreadStarten( &Filter );
 	
 	// Die Hinrichtung luft hier im Hauptthread ab
	GuiNachEngine( &Filter );

	// Handles schlieen
	CloseHandle( Filter.hEngineNachFilter );
	CloseHandle( Filter.hVonEngine );
	CloseHandle( Filter.hNachEngine );

	// Fertig
	return 0;
}

